/**
 * Original HTTP Server sourcecode:
 * (c) 2001 Sam Pullara  sam@sampullara.com
 * Updated HTTP Server sourcecode+multicast additions:
 * (c) 2006 Daniel Kirstenpfad btk@technology-ninja.com
 */

using System;
using System.Collections;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Text;

namespace YAPS
{
    #region HttpProcessor
    /// <summary>
    /// Implements a Handler for each HTTP Client Request
    /// </summary>
    public class HttpProcessor
    {
        #region Variables'r'us
        private static int threads = 0;
		private Socket s;
        public HttpServer HTTPServer;
		public NetworkStream ns;
		private StreamReader sr;
		private StreamWriter sw;
		private string method;
		private string url;
        private string querystring;
		private string protocol;
		private Hashtable headers;
		private string request;
		private bool keepAlive = false;
		private int numRequests = 0;
		private bool verbose = HttpServer.verbose;
		private byte[] bytes = new byte[4096];
		private FileInfo docRootFile;
        #endregion

        #region Constructor
        /// <summary>
        /// Each HTTP processor object handles one client.  If Keep-Alive is enabled then this
        /// object will be reused for subsequent requests until the client breaks keep-alive.
        /// This usually happens when it times out.  Because this could easily lead to a DoS
        /// attack, we keep track of the number of open processors and only allow 100 to be
        /// persistent active at any one time.  Additionally, we do not allow more than 500
        /// outstanding requests.
        /// </summary>
        /// <param name="docRoot">Root-Directory of the HTTP Server</param>
        /// <param name="s">the Socket to work with</param>
        /// <param name="webserver">the "master" HttpServer Object of this Client</param>
		public HttpProcessor(string docRoot, Socket s, HttpServer webserver) {
			this.s = s;
			docRootFile = new FileInfo(docRoot);
			headers = new Hashtable();
            HTTPServer = webserver;
        }
        #endregion

        #region processor starting point
        /// <summary>
        /// This is the main method of each thread of HTTP processing.  We pass this method
        /// to the thread constructor when starting a new connection.
        /// </summary>
		public void process() {
			try 
			{
                // Increment the number of current connections
				Interlocked.Increment(ref threads);
				// Bundle up our sockets nice and tight in various streams
				ns = new NetworkStream(s, FileAccess.ReadWrite);
				// It looks like these streams buffer
				sr = new StreamReader(ns);
				sw = new StreamWriter(ns);
				// Parse the request, if that succeeds, read the headers, if that
				// succeeds, then write the given URL to the stream, if possible.
				while (parseRequest())
				{
					if (readHeaders()) {
						// This makes sure we don't have too many persistent connections and also
						// checks to see if the client can maintain keep-alive, if so then we will
						// keep this http processor around to process again.
						if (threads <= 100 && "Keep-Alive".Equals(headers["Connection"]))
						{
							keepAlive = true;
						}
						// Copy the file to the socket
						writeURL();
						// If keep alive is not active then we want to close down the streams
						// and shutdown the socket
						if (!keepAlive)
						{
							ns.Close();
                            try
                            {
                                s.Shutdown(SocketShutdown.Both);
                            }
                            catch (Exception) { }
							break;
						}
					}
				}
			} finally {
				// Always decrement the number of connections
				Interlocked.Decrement(ref threads);	
			}
        }
        #endregion

        #region Validationchecks
        /// <summary>
        /// parses the Request and determines if it's actually a valid one or not
        /// </summary>
        /// <returns>is valid request(true) or not(false)</returns>
        public bool parseRequest() {
			// The number of requests handled by this persistent connection
			numRequests++;
			// Here is where we ensure that we are not overloaded
			if (threads > 500) {
				writeError(502, "Server temporarily overloaded");
				return false;
			}
			// FIXME: This could conceivably used to DoS us if we never finish reading the
			// line and they never hang up.  We could set the socket options to limit
			// the amount of time before reading a request.
			try 
			{
				request = null;
				request = sr.ReadLine();
			} 
			catch (IOException) 
			{
			}
			// If the request line is null, then the other end has hung up on us.  A well
			// behaved client will do this after 15-60 seconds of inactivity.
			if (request == null) {
				if (verbose) {
					ConsoleOutputLogger.WriteLine("Keep-alive broken after " + numRequests + " requests");
				}
				return false;
			}
			// HTTP request lines are of the form:
			// [METHOD] [Encoded URL] HTTP/1.?
			string[] tokens = request.Split(new char[]{' '});
			if (tokens.Length != 3) {
				writeError(400, "Bad request");
				return false;
			}
			// We currently only handle GET requests
			method = tokens[0];
			if(!method.Equals("GET")) {
				writeError(501, method + " not implemented");
				return false;
			}
			url = tokens[1];
			// Only accept valid urls
			if (!url.StartsWith("/")) {
				writeError(400, "Bad URL");
				return false;
			}
			// Decode all encoded parts of the URL using the built in URI processing class
			int i = 0;
			while((i = url.IndexOf("%", i)) != -1)
			{
				url = url.Substring(0, i) + Uri.HexUnescape(url, ref i) + url.Substring(i);
			}
			// Lets just make sure we are using HTTP, thats about all I care about
			protocol = tokens[2];
			if (!protocol.StartsWith("HTTP/")) {
				writeError(400, "Bad protocol: " + protocol);
			}
			return true;
		}

        /// <summary>
        /// Reads the headers of the request and determines if they are valid or not
        /// </summary>
        /// <returns>is valid request(true) or not(false)</returns>
		public bool readHeaders() {
			string line;
			string name = null;
			// The headers end with either a socket close (!) or an empty line
			while((line = sr.ReadLine()) != null && line != "") {
				// If the value begins with a space or a hard tab then this
				// is an extension of the value of the previous header and
				// should be appended
				if (name != null && Char.IsWhiteSpace(line[0])) {
					headers[name] += line;
					continue;
				}
				// Headers consist of [NAME]: [VALUE] + possible extension lines
				int firstColon = line.IndexOf(":");
				if (firstColon != -1) {
					name = line.Substring(0, firstColon);
					String value = line.Substring(firstColon + 1).Trim();
					if (verbose) ConsoleOutputLogger.WriteLine(name + ": " + value);
					headers[name] = value;
				} else {
					writeError(400, "Bad header: " + line);
					return false;
				}
			}
			return line != null;
        }
        #endregion

        #region HTML OK Message Generator
        String ForwardToIndexHTML(String Message)
        {
            String Output = "";

            Output = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"><html><head><title>"+Message+"</title><meta http-equiv=\"refresh\" content=\"5; URL=/index.html\"></head><body><a href=\"/index.html\">"+Message+"</a></body></html>";

            return Output;
        }
        #endregion

        #region Request URL handling
        /// <summary>
        /// We need to make sure that the url that we are trying to treat as a file
        /// lies below the document root of the http server so that people can't grab
        /// random files off your computer while this is running.
        /// </summary>
		public void writeURL() {
			try {
                querystring = "";
                ConsoleOutputLogger.WriteLine("Request: " + url);
                if (url.StartsWith("/vcr/"))
                {
                    #region VCR request
                    // remove the /vcr/ stuff
                    url = url.Remove(0,5);

                    // check which function is called

                    #region AddRecording
                    if (url.ToUpper().StartsWith("ADDRECORDING"))
                    {
                        url = url.Remove(0, 13);

                        string[] parameterlist = url.Split('&');

                        // instantiate new Recording
                        Recording newTimer = new Recording();


                        #region data'n'stuff
                        int s_date = 0, s_month = 0, s_year = 0, s_hour = 0, s_minute = 0, e_date = 0, e_month = 0, e_year = 0, e_hour = 0, e_minute = 0;
                        #endregion

                        // TODO: add some try..catch thingies..
                        foreach (string Parameter in parameterlist)
                        {
                            try
                            {
                                string[] split_parameter = Parameter.Split('=');

                                // Zwischenergebniss: s_date=1&s_month=8&s_year=2006&s_hour=12&s_minute=12&e_date=1
                                // &e_month=8&e_year=2006&e_hour=12&e_minute=12&name=1asdfasdfasf&ch=25

                                switch (split_parameter[0])
                                {
                                    case "s_date":
                                        s_date = Convert.ToInt32(split_parameter[1]);
                                        break;
                                    case "s_month":
                                        s_month = Convert.ToInt32(split_parameter[1]);
                                        break;
                                    case "s_year":
                                        s_year = Convert.ToInt32(split_parameter[1]);
                                        break;
                                    case "s_hour":
                                        s_hour = Convert.ToInt32(split_parameter[1]);
                                        break;
                                    case "s_minute":
                                        s_minute = Convert.ToInt32(split_parameter[1]);
                                        break;
                                    case "e_date":
                                        e_date = Convert.ToInt32(split_parameter[1]);
                                        break;
                                    case "e_month":
                                        e_month = Convert.ToInt32(split_parameter[1]);
                                        break;
                                    case "e_year":
                                        e_year = Convert.ToInt32(split_parameter[1]);
                                        break;
                                    case "e_hour":
                                        e_hour = Convert.ToInt32(split_parameter[1]);
                                        break;
                                    case "e_minute":
                                        e_minute = Convert.ToInt32(split_parameter[1]);
                                        break;
                                    case "name":
                                        newTimer.Recording_Name = split_parameter[1];
                                        break;
                                    case "ch":
                                        newTimer.Channel = split_parameter[1];
                                        break;
                                    default:
                                        // alles andere interessiert uns nicht
                                        break;
                                }
                            }
                            catch (Exception e)
                            {
                                ConsoleOutputLogger.WriteLine("AddRecording: "+e.Message);
                            }
                        }
                        // TODO: check if there are enough information given to actually create that timer
                        try
                        {
                            newTimer.StartsAt = new DateTime(s_year, s_month, s_date, s_hour, s_minute, 0);
                            newTimer.EndsAt = new DateTime(e_year, e_month, e_date, e_hour, e_minute, 0);
                        }
                        catch (Exception e)
                        {
                            ConsoleOutputLogger.WriteLine("AddRecording: " + e.Message);
                        }
                        newTimer.Recording_Filename = newTimer.Recording_ID.ToString();

                        // TODO: check for other timers that could possibly concur...
                        lock (HTTPServer.vcr_scheduler.Recordings.SyncRoot)
                        {
                            HTTPServer.vcr_scheduler.Recordings.Add(newTimer.Recording_ID, newTimer);
                        }

                        // tell the client that everything went fine...
                        string Output = ForwardToIndexHTML("New Timer created...");
                        byte[] buffer = new UnicodeEncoding().GetBytes(Output);
                        int left = new UnicodeEncoding().GetByteCount(Output);
                        writeSuccess(left);
                        ns.Write(buffer, 0, left);
                        ns.Flush();

                        // Save the new Configuration...
                        HTTPServer.Configuration.SaveSettings();
                    #endregion
                    }
                    else
                    if (url.ToUpper().StartsWith("DELETERECORDING"))
                    {
                    #region DeleteRecording

                        // remove the /vcr/deleterecording? stuff
                        url = url.Remove(0, 16);

                        string[] splitted = url.Split('=');
                        if (splitted[0].ToUpper() == "ID")
                        {
                            Guid delete_entry_id = Guid.NewGuid();
                            string recording_name = "";

                            lock (HTTPServer.vcr_scheduler.Recordings.SyncRoot)
                            {
                                foreach (Recording recording_entry in HTTPServer.vcr_scheduler.Recordings.Values)
                                {
                                    string original = Convert.ToBase64String(recording_entry.Recording_ID.ToByteArray());

                                    if (original == (splitted[1]+"=="))
                                    {
                                        recording_name = recording_entry.Recording_Name;
                                        delete_entry_id = recording_entry.Recording_ID;
                                    }
                                }
                            }

                            bool deleted = true;
                            try
                            {
                                lock (HTTPServer.vcr_scheduler.Recordings.SyncRoot)
                                {
                                    HTTPServer.vcr_scheduler.Recordings.Remove(delete_entry_id);
                                }
                            }
                            catch (Exception e)
                            {
                                // failed
                                deleted = false;
                            }
                            // tell the client that everything went fine...
                            string Output;
                            if (deleted)
                                Output = ForwardToIndexHTML("Recording: "+recording_name+" successfully deleted...");
                            else
                                Output = ForwardToIndexHTML("Error: Recording not deleted...");
                            byte[] buffer = new UnicodeEncoding().GetBytes(Output);
                            int left = new UnicodeEncoding().GetByteCount(Output);
                            writeSuccess(left);
                            ns.Write(buffer, 0, left);
                            ns.Flush();

                            // Save the new Configuration...
                            HTTPServer.Configuration.SaveSettings();
                        }
                        else
                            writeFailure();

                        #endregion
                    }
                    else
                    if (url.ToUpper().StartsWith("REMOVERECORDINGFILE"))
                    {
                    #region RemoveRecordingFile

                        // remove the deleterecording? stuff
                        url = url.Remove(0, 20);

                        string[] splitted = url.Split('=');
                        if (splitted[0].ToUpper() == "ID")
                        {
                            Guid delete_entry_id = Guid.NewGuid();
                            String delete_entry_filename = "";

                            string recording_name = "";

                            lock (HTTPServer.vcr_scheduler.doneRecordings.SyncRoot)
                            {
                                foreach (Recording recording_entry in HTTPServer.vcr_scheduler.doneRecordings.Values)
                                {
                                    string original = recording_entry.Recording_ID.ToString();

                                    if (original == splitted[1])
                                    {
                                        recording_name = recording_entry.Recording_Name;
                                        delete_entry_id = recording_entry.Recording_ID;
                                        delete_entry_filename = recording_entry.Recording_Filename;
                                    }
                                }
                            }
                            bool deleted = true;
                            try
                            {
                                lock (HTTPServer.vcr_scheduler.doneRecordings.SyncRoot)
                                {
                                    // remove file
                                    File.Delete(delete_entry_filename);

                                    // remove entry in Hashtable
                                    HTTPServer.vcr_scheduler.doneRecordings.Remove(delete_entry_id);
                                }
                            }
                            catch (Exception e)
                            {
                                // failed
                                deleted = false;
                            }
                            // tell the client that everything went fine...
                            string Output;
                            if (deleted)
                                Output = ForwardToIndexHTML("Recording: " + recording_name + " successfully deleted...");
                            else
                                Output = ForwardToIndexHTML("Error: Recording not deleted...");
                            byte[] buffer = new UnicodeEncoding().GetBytes(Output);
                            int left = new UnicodeEncoding().GetByteCount(Output);
                            writeSuccess(left);
                            ns.Write(buffer, 0, left);
                            ns.Flush();

                            // Save the new Configuration...
                            HTTPServer.Configuration.SaveSettings();
                        }
                        else
                            writeFailure();

                        #endregion
                    }

                    else
                    {
                        // nothing to do...
                        writeError(404, "No Method found");
                    }
                    #endregion
                }
                else
                if (url.StartsWith("/dvb/"))
                {
                    #region Streaming Request
                    // check if there is a MulticastProcessor Object for this channel
                    VCRandStreaming HReq = new VCRandStreaming(false,null);

                    HReq.HandleStreaming(url, this);

                    while (!HReq.done)
                    {
                        Thread.Sleep(100);
                    }
                    #endregion
                }
                else
                {
                    #region File request (everything else...)

                    // check if we have some querystring parameters
                    if (url.Contains("?"))
                    {
                        // yes, remove everything after the ? from the url but save it to querystring
                        querystring = url.Substring(url.IndexOf('?')+1);
                        url = url.Remove(url.IndexOf('?'));
                    }
                    // no... nothing to do here...

                    // Replace the forward slashes with back-slashes to make a file name
                    string filename = url.Replace('/', '\\');
                    // Construct a filename from the doc root and the filename
                    FileInfo file = new FileInfo(docRootFile + filename);
                    // Make sure they aren't trying in funny business by checking that the
                    // resulting canonical name of the file has the doc root as a subset.
                    filename = file.FullName;
                    if (!filename.StartsWith(docRootFile.FullName))
                    {
                        writeForbidden();
                    }
                    else
                    {
                        FileStream fs = null;
                        BufferedStream bs = null;

                        try
                        {
                            if (filename.EndsWith(".html")|(filename.EndsWith(".htm")))
                            {
                                String Output = HTTPServer.Template_Processor.ProcessHTMLTemplate(filename,querystring);

                                int left = new UnicodeEncoding().GetByteCount(Output);
                                
                                writeSuccess(left);

                                byte[] buffer = new UnicodeEncoding().GetBytes(Output);

                                ns.Write(buffer, 0, left);

                                ns.Flush();

                            }
                            else
                            {
                                // Open the file for binary transfer
                                fs = new FileStream(filename, FileMode.Open, FileAccess.Read);
                                // Write the content length and the success header to the stream
                                long left = file.Length;
                                writeSuccess(left);
                                // Copy the contents of the file to the stream, ensure that we never write
                                // more than the content length we specified.  Just in case the file somehow
                                // changes out from under us, although I don't know if that is possible.
                                bs = new BufferedStream(fs);
                                int read;
                                while (left > 0 && (read = bs.Read(bytes, 0, (int)Math.Min(left, bytes.Length))) != 0)
                                {
                                    ns.Write(bytes, 0, read);
                                    left -= read;
                                }
                                ns.Flush();
                                bs.Close();
                                fs.Close();
                            }
                        }
                        catch (Exception e)
                        {
                            ConsoleOutputLogger.WriteLine(e.Message);
                            if (bs!=null) bs.Close();
                            if (bs!=null) fs.Close();
                        }

                    }
                    #endregion
                }
			} catch(FileNotFoundException) {
				writeFailure();
			}
        }
        #endregion

        #region Simple HTTP Responses+Codes
        /**
		 * These write out the various HTTP responses that are possible with this
		 * very simple web server.
		 */
		public void writeSuccess(long length) {
			writeResult(200, "OK", length);
		}

		public void writeFailure() {
			writeError(404, "File not found");
		}
	
		public void writeForbidden() {
			writeError(403, "Forbidden");
		}

		public void writeError(int status, string message) 
		{
			string output = "<h1>HTTP/1.0 " + status + " " + message + "</h1>";
			writeResult(status, message, (long) output.Length);
            ConsoleOutputLogger.WriteLine("Error " + status);
			sw.Write(output);
			sw.Flush();
		}

		public void writeResult(int status, string message, long length) 
		{
			if (verbose) ConsoleOutputLogger.WriteLine(request + " " + status + " " + numRequests);
			sw.Write("HTTP/1.0 " + status + " " + message + "\r\n");
			sw.Write("Content-Length: " + length + "\r\n");
			if (keepAlive)
			{
				sw.Write("Connection: Keep-Alive\r\n");
			}
			else
			{
				sw.Write("Connection: close\r\n");
			}
			sw.Write("\r\n");
			sw.Flush();
        }
        #endregion
    }
    #endregion

    #region HttpServer
    /// <summary>
    /// Implements a HTTP Server Listener
    /// </summary>
    public class HttpServer
    {
        #region Data
        public Settings Configuration;
        public Hashtable MulticastProcessorList;
		public static bool verbose = false;
		private int port;
		private string docRoot;
        private VCRScheduler internal_vcr_scheduler;
        private bool internal_vcr_scheduler_set;
        public TemplateProcessor Template_Processor;

        public VCRScheduler vcr_scheduler
        {
            get
            {
                return internal_vcr_scheduler;
            }
            set
            {
                internal_vcr_scheduler = value;
                internal_vcr_scheduler_set = true;
            }
        }

        #endregion

        #region Construction
        public HttpServer(string docRoot, int port) {
			this.docRoot = docRoot;
			this.port = port;
            MulticastProcessorList = new Hashtable();
            internal_vcr_scheduler_set = false;
        }
        #endregion

        #region Listener
        /// <summary>
        /// Create a new server socket, set up all the endpoints, bind the socket and then listen
        /// </summary>
        public void listen() 
        {
            // Wait for VCRScheduler...
            ConsoleOutputLogger.WriteLine("Waiting for VCRScheduler...");
            while (!internal_vcr_scheduler_set)
            {
                Thread.Sleep(10);
            }
            // create the Template Processor
            Template_Processor = new TemplateProcessor(internal_vcr_scheduler);


			// Create a new server socket, set up all the endpoints, bind the socket and then listen
			Socket listener = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
			IPAddress ipaddress = IPAddress.Parse("0.0.0.0");
			IPEndPoint endpoint = new IPEndPoint(ipaddress, port);
			listener.Bind(endpoint);
			listener.Blocking = true;
			listener.Listen(-1);
			ConsoleOutputLogger.WriteLine("Http server listening on port " + port);
			while(true) {
				try {
					// Accept a new connection from the net, blocking till one comes in
					Socket s = listener.Accept();
					// Create a new processor for this request
					HttpProcessor processor = new HttpProcessor(docRoot, s, this);
					
                    // changed from multithreaded to single threaded due to port binding problems
                    /// TODO: Port binding problems
                    /// 
                    // Dispatch that processor in its own thread
					Thread thread = new Thread(new ThreadStart(processor.process));
					thread.Start();
                    Thread.Sleep(10);
                    //processor.process();

				} catch(NullReferenceException) {
					// Don't even ask me why they throw this exception when this happens
					ConsoleOutputLogger.WriteLine("Accept failed.  Another process might be bound to port " + port);
				}
			}
        }
        #endregion
    }
    #endregion
}

